import json
import requests
import logging
from urllib.parse import urljoin

from ..exceptions import AuthenticationError, ResourceNotFoundError, InternalServerError, ConnectionError
from .. import settings

requests_log = logging.getLogger("requests")

if settings.REQUESTS_LOG_LEVEL == 'warning':
    requests_log.setLevel(logging.WARNING)
elif settings.REQUESTS_LOG_LEVEL == 'debug':
    requests_log.setLevel(logging.DEBUG)
elif settings.REQUESTS_LOG_LEVEL == 'info':
    requests_log.setLevel(logging.INFO)

JSON_HTTP_HEADERS = {'content-type': 'application/json'}

class RequestsMixin:
    def __init__(self):
        self.entity_id = ""
        self.api_key = ""
        self.session = requests.Session()
        self.connected = False
        
        self._add_hooks('pre_connect', 'post_connect', 'pre_disconnect', 'post_disconnect')


    def connect(self):
        """
        Register a connection with the HPIT Server.

        This essentially sets up a session and logs that you are actively using
        the system. This is mostly used to track plugin use with the site.
        """
        self._try_hook('pre_connect')

        connection = self._post_data('connect', {
                'entity_id': self.entity_id, 
                'api_key': self.api_key
            }
        )

        self.connected = True
        self._try_hook('post_connect')
       
        return self.connected


    def disconnect(self):
        """
        Tells the HPIT Server that you are not currently going to poll
        the server for messages or responses. This also destroys the current session
        with the HPIT server.
        """
        self._try_hook('pre_disconnect')
        
        self._post_data('disconnect', {
                'entity_id': self.entity_id,
                'api_key': self.api_key
            }
        )

        self.connected = False
        self._try_hook('post_disconnect')

        return self.connected


    def _post_data(self, url, data=None):
        """
        Sends arbitrary data to the HPIT server. This is mainly a thin
        wrapper ontop of requests that ensures we are using sessions properly.

        Returns: requests.Response : class - The response from HPIT. Normally a 200:OK.
        """
        url = urljoin(settings.HPIT_URL_ROOT, url)

        failure_count = 0
        while failure_count < 3:
            try:
                if data:
                    response = self.session.post(url, data=json.dumps(data), headers=JSON_HTTP_HEADERS)
                else:
                    response = self.session.post(url)
                break

            except requests.exceptions.ConnectionError:
                if failure_count == 3:
                    raise ConnectionError("Could not connect to server. Tried 3 times.")

                failure_count += 1
                continue

            except Exception:
                raise ConnectionError("The remote address is bogus.")
        
        if response.status_code == 200:
            return response
        elif response.status_code == 403:
            raise AuthenticationError("Request could not be authenticated")
        elif response.status_code == 404:
            raise ResourceNotFoundError("Requested resource not found")
        elif response.status_code == 500:
            raise InternalServerError("Internal server error")


    def _get_data(self, url):
        """
        Gets arbitrary data from the HPIT server. This is mainly a thin
        wrapper on top of requests that ensures we are using session properly.

        Returns: dict() - A Python dictionary representing the JSON recieved in the request.
        """
        url = urljoin(settings.HPIT_URL_ROOT, url)

        failure_count = 0
        while failure_count < 3:
            try:
                response = self.session.get(url)
                break

            except requests.exceptions.ConnectionError as e:
                if failure_count == 3:
                    raise e

                failure_count += 1
                continue

        if response.status_code == 200:
            return response.json()
        elif response.status_code == 403:
            raise AuthenticationError("Request could not be authenticated")
        elif response.status_code == 404:
            raise ResourceNotFoundError("Requested resource not found")
        elif response.status_code == 500:
            raise InternalServerError("Internal server error")


    def send_log_entry(self, text):
        """
        Send a log entry to the HPIT server.
        """
        self._post_data("log", data={'log_entry':text})


    def _add_hooks(self, *hooks):
        """
        Adds hooks to this class. If the function is already defined, this leaves that definition. If 
        it doesn't exists the hook is created and set to None
        """
        for hook in hooks:
            if not hasattr(self, hook):
                setattr(self, hook, None)


    def _try_hook(self, hook_name):
        """
        Try's to call a signal hook. Hooks take in no parameters and return a boolean result.
        True will cause the plugin to continue execution.
        False will cause the plugon to stop execution.
        """
        hook = getattr(self, hook_name, None)

        if hook:
            return hook()
        else:
            return True




import time

from .message_sender_mixin import MessageSenderMixin
from .exceptions import PluginPollError, BadCallbackException

class Plugin(MessageSenderMixin):
    def __init__(self, entity_id, api_key, wildcard_callback=None):
        super().__init__()

        self.run_loop = True
        self.entity_id = str(entity_id)
        self.api_key = str(api_key)
        self.wildcard_callback = wildcard_callback
        self.transaction_callback = None
        self.callbacks = {}

        self.poll_wait = 500
        self.time_last_poll = time.time() * 1000

        self._add_hooks(
            'pre_poll_messages', 'post_poll_messages', 
            'pre_dispatch_messages', 'post_dispatch_messages', 
            'pre_handle_transactions', 'post_handle_transcations')


    def post_connect(self):
        pass


    def register_transaction_callback(self,callback):
        """
        Set a callback for transactions and start listening for them.
        """
        if not hasattr(callback,"__call__"):
            raise BadCallbackException("The callback submitted is not callable.")
        self._post_data('plugin/subscribe', {'message_name' : "transaction"})
        self.transaction_callback = callback
        
        
    def clear_transaction_callback(self):
        """
        Clear the callback for transactions and stop listening for them.
        """
        self._post_data('plugin/unsubscribe', {'message_name': "transaction"})
        self.transaction_callback = None
        

    def list_subscriptions(self):
        """
        Polls the HPIT server for a list of message names we currently subscribing to.
        """
        subscriptions = self._get_data('plugin/subscription/list')['subscriptions']

        for sub in subscriptions:
            if sub not in self.callbacks:
                self.callbacks[sub] = None

        return self.callbacks


    def subscribe(self, **messages):
        """
        Subscribe to messages, each argument is exepcted as a key value pair where
        the key is the message's name and the value is the callback function.
        """
        for message_name, callback in messages.items():
            self._post_data('plugin/subscribe', {'message_name' : message_name})
            self.callbacks[message_name] = callback
            

    def unsubscribe(self, *message_names):
        """
        Unsubscribe from messages. Pass each message name as a separate parameter.
        """
        for message_name in message_names:
            if message_name in self.callbacks:
                self._post_data('plugin/unsubscribe', {'message_name': message_name})
                del self.callbacks[message_name]


    def _poll(self):
        """
        Get a list of new messages from the server for messages we are listening 
        to.
        """
        return self._get_data('plugin/message/list')['messages']

    def _handle_transactions(self):
        """
        Get a list of datashop transactions from the server. 
        """

        transaction_data = self._get_data('plugin/transaction/list')['transactions']

        for item in transaction_data:
            if self.transaction_callback:
                if not self.transaction_callback(item):
                    return False

        return True


    def _dispatch(self, message_data):
        """
        For each message recieved route it to the appropriate callback.
        """
        if not self._try_hook('pre_dispatch_messages'):
            return False

        for message_item in message_data:
            message = message_item['message_name']
            payload = message_item['message']

            #Inject the message_id into the payload
            payload['message_id'] = message_item['message_id']
            payload['sender_entity_id'] = message_item['sender_entity_id']

            try:
                self.callbacks[message](payload)
            except KeyError:
                #No callback registered try the wildcard
                if self.wildcard_callback:
                    try:
                        self.wildcard_callback(payload)
                    except TypeError:
                        raise PluginPollError("Wildcard Callback is not a callable")
            except TypeError as e:
                #Callback isn't a function
                if self.callbacks[message] is None:
                    raise PluginPollError("No callback registered for message: <" + message + ">")
                else:
                    raise e

        if not self._try_hook('post_dispatch_messages'):
            return False

        return True


    def start(self):
        """
        Start the plugin. Connect to the HPIT server. Then being polling and dispatching
        message callbacks based on messages we subscribe to.
        """
        self.connect()
        self.list_subscriptions()

        try:
            while self.run_loop:

                #A better timer
                cur_time = time.time() * 1000

                if cur_time - self.time_last_poll < self.poll_wait:
                    continue;

                self.time_last_poll = cur_time

                #Handle messages submitted by tutors
                if not self._try_hook('pre_poll_messages'):
                    break;

                message_data = self._poll()

                if not self._try_hook('post_poll_messages'):
                    break;

                if not self._dispatch(message_data):
                    return False

                if not self._try_hook('pre_handle_transactions'):
                    break;

                if not self._handle_transactions():
                    return False

                if not self._try_hook('post_handle_transactions'):
                    break;

                #Handle responses from other plugins
                if not self._try_hook('pre_poll_responses'):
                    break;

                responses = self._poll_responses()

                if not self._try_hook('post_poll_responses'):
                    break;

                if not self._dispatch_responses(responses):
                    break;

        except KeyboardInterrupt:
            pass

        self.disconnect()


    def stop(self):
        self.run_loop = False


    def send_response(self, message_id, payload):
        """
        Sends a response to HPIT upon handling a specific message.

        Responses are handled differently than normal messages as they are destined
        for a only the original sender of the message_id to recieve the response.
        """
        self._post_data('response', {
            'message_id': message_id,
            'payload': payload
        })



class ConnectionError(Exception):
    """
    This exception indicates a generic HPIT connection problem.
    """

class AuthenticationError(Exception):
    """
    This exception raised on HPIT 403.
    """
    pass

class ResourceNotFoundError(Exception):
    """
    This exception raised on HPIT 403.
    """
    pass

class InternalServerError(Exception):
    """
    This exception raised on HPIT 500.
    """
    pass

class PluginRegistrationError(Exception):
    """
    This exception indicates that a plugin could not register with HPIT.
    """
    pass

class PluginPollError(Exception):
    """
    This exception indicates that a plugin could not poll HPIT.
    """
    pass

class ResponseDispatchError(Exception):
    """
    This exception indicates that a response from HPIT could not be dispatched to a callback.
    """
    pass

class InvalidMessageNameException(Exception):
    """
    This exception is raised when a user attempts to use a system message name, like 'transaction'.
    """
    
class BadCallbackException(Exception):
    """
    Raised when a callback is not callable
    """
    



class HpitClientSettings:
    instance = None
    HPIT_URL_ROOT = 'https://www.hpit-project.org'
    REQUESTS_LOG_LEVEL = 'debug'

    def __new__(cls, *args, **kwargs):
        if cls.instance:
            raise Exception("Settings has already been created.")

        cls.instance = object.__new__(cls)
        cls.instance.__init__(*args, **kwargs)
        return cls.instance

    @classmethod
    def settings(cls):
        if not cls.instance:
            cls.instance = HpitClientSettings()

        return cls.instance



from .requests_mixin import RequestsMixin
from ..exceptions import ResponseDispatchError
from ..exceptions import InvalidMessageNameException

class MessageSenderMixin(RequestsMixin):
    def __init__(self):
        super().__init__()
        self.response_callbacks = {}
        
        self._add_hooks('pre_poll_responses', 'post_poll_responses', 'pre_dispatch_responses', 'post_dispatch_responses')

    def send(self, message_name, payload, callback=None):
        """
        Sends a message to the HPIT server. Messages are the meat of how
        HPIT works. All messages are asyncronous and non-blocking. Responses
        are not guarenteed to be prompt due to the distributed nature of HPIT. However,
        it is our goal to ensure response times are as minimal as possible. Responses times
        are dependendent on the network, HPIT server load, and the networking and processing
        constraints of the plugins.

        Each message consists of two main things:
            - message_name: string - The name of the message we are sending to HPIT. This will determine
            which plugins recieves the message.
            - payload: dict - A Python dictionary of the data to send to HPIT and to whatever plugins
            are listening to and will process the message. The dictionary will be converted to JSON in
            route to HPIT.

        Optionally you can pass a callback, and as this message sender polls HPIT for responses
        !!!IF!!! it recieved such a response from a plugin your callback will be called to handle
        the response with any information from the plugin.

        Returns: requests.Response : class - A request.Response object returned from submission 
        of the message. This is not the eventual response from HPIT. It is simply an acknowledgement
        the data was recieved. You must send in a callback to handle the actual HPIT response.
        """
        
        if message_name == "transaction":
            raise InvalidMessageNameException("Cannot use message_name 'transaction'.  Use send_transaction() method for datashop transactions.")
            
        response = self._post_data('message', {
            'name': message_name,
            'payload': payload
        }).json()

        if callback:
            self.response_callbacks[response['message_id']] = callback

        return response
        
    def send_transaction(self, payload, callback= None):
        """
        This method functions identially as send, but inserts "transaction" as the message.
        This is specifically for DataShop transactions.
        See send() method for more details.
        """
        
        response = self._post_data('transaction', {
            'payload': payload
        }).json()

        if callback:
            self.response_callbacks[response['message_id']] = callback

        return response
        


    def _poll_responses(self):
        """
        This function polls HPIT for responses to messages we submitted earlier on.

        Hooks:
            self.pre_poll_responses - If set to a callable, will be called before the poll
            request is made to HPIT.
            self.post_poll_responses - If set to a callable, will be called after the poll
            request is made to HPIT.

        Returns: dict - The list of responses from the server for earlier messages 
        submitted by this message sender to HPIT.
        """
        if not self._try_hook('pre_poll_responses'):
            return False

        responses = self._get_data('response/list')['responses']

        if not self._try_hook('post_poll_responses'):
            return False

        return responses


    def _dispatch_responses(self, responses):
        """
        This function is responsible for dispatching responses to earlier message to
        their callbacks that were set when the transcation was sent with self.send.

        Hooks:
            self.pre_dispatch_responses - If set to a callable, will be called before the 
            responses are dispatched to their respective callbacks.
            self.post_dispatch_responses - If set to a callable, will be called after the
            responses are dispatched to their respective callbacks.

        Returns: boolean - True if event loop should continue. False if event loop should 
            abort.
        """
        if not self._try_hook('pre_dispatch_responses'):
            return False

        for res in responses:
            try:
                message_id = res['message']['message_id']
            except KeyError:
                raise ResponseDispatchError('Invalid response from HPIT. No message id supplied in response.')

            try:
                response_payload = res['response']
            except KeyError:
                raise ResponseDispatchError('Invalid response from HPIT. No response payload supplied.')

            try:
                self.response_callbacks[message_id](response_payload)
            except KeyError:
                raise ResponseDispatchError('No callback registered for message id: ' + message_id)
            except TypeError:
                raise ResponseDispatchError("Callback registered for transcation id: " + message_id + " is not a callable.")

        if not self._try_hook('post_dispatch_responses'):
            return False

        return True



import time

from .message_sender_mixin import MessageSenderMixin
from .exceptions import ResponseDispatchError

class Tutor(MessageSenderMixin):
    def __init__(self, entity_id, api_key, callback, **kwargs):
        super().__init__()

        self.run_loop = True
        self.entity_id = str(entity_id)
        self.api_key = str(api_key)
        self.callback = callback

        self.poll_wait = 500
        self.time_last_poll = time.time() * 1000

        for k, v in kwargs.items():
            setattr(self, k, v)


    def start(self):
        """
        Starts the tutor in event-driven mode.
        """
        self.connect()
        
        try:
            while self.run_loop:
                if not self.callback():
                    break;

                #A better timer
                cur_time = time.time() * 1000

                if cur_time - self.time_last_poll < self.poll_wait:
                    continue;

                self.time_last_poll = cur_time

                if not self._try_hook('pre_poll_responses'):
                    break;

                responses = self._poll_responses()

                if not self._try_hook('post_poll_responses'):
                    break;

                try:
                    if not self._dispatch_responses(responses):
                        break;
                except ResponseDispatchError as e:
                    print(str(e))

        except KeyboardInterrupt:
            pass

        self.disconnect()


    def stop(self):
        self.run_loop = False
